'''
LMSworker  - queue tracks using musly
(c) 2017 R.S.U. / GPL v3 / https://www.nexus0.net/pub/sw/lmsmusly
'''

import threading
import queue
import logging, os, time, collections, datetime

from lmsmusly.lms.server import LMSServer
from lmsmusly.utils import config

_LOGGER = logging.getLogger(__name__)

class LMSWorker():

	def worker( self, event, q, lms, mus, mta):
		#_LOGGER.debug( "LMSWorker:worker enter")
#		global main_run
		run_worker = True
		tracks_served = collections.deque( maxlen=config.MSKIPDUPS_LEN)
		seedtracks_used = []
		while run_worker:
			#_LOGGER.debug( "LMSWorker:worker enter loop")
			event.wait()
			_LOGGER.debug( "LMSWorker:worker: run {}".format( run_worker))
			try:
				msg =q.get(block=False)
				_LOGGER.debug( "LMSWorker:worker: msg: {}".format( repr(msg)))
				if msg == "exit":
					run_worker = False
				elif msg == "clear_duptracks":
					_LOGGER.debug( "LMSWorker:worker: command clear_duptracks")
				q.task_done()
			except queue.Empty:
				pass

			lms.update()
			if not lms.is_player_known( config.LMS_PLAYER):
				_LOGGER.warn( "LMSWorker:worker: player with id {} not found".format( config.LMS_PLAYER))
				time.sleep( config.LMS_POLLINTERVAL)
				continue
			player = lms.player( config.LMS_PLAYER)
			curr_idx = player.get_current_index()
			pls_len = player.get_playlist_len()
			_LOGGER.debug( "LMSWorker: is_playing= {} curr_idx={} plslen={}".format( player.is_playing(), curr_idx, pls_len))

			if curr_idx == pls_len-1:
				seedtrack = player.get_track_fileurl( config.LMS_SEEDTRACKIDX)
				if seedtrack == None:
					_LOGGER.warn("LMSWorker: seed track #{} not found".format(config.LMS_SEEDTRACKIDX))
					for i in range( max( config.LMS_SEEDTRACKIDX-1, 0), -1, -1):
						seedtrack = player.get_track_fileurl( i)
						if seedtrack != None:
							break
				if seedtrack == None:
					_LOGGER.warn("LMSWorker: seed track #{} not found, failed to find alternative".format(config.LMS_SEEDTRACKIDX))
					mus.set_stat( 'lastappmsg', '<span class="text-danger">Error:</span> seed track #{} not found'.format( config.LMS_SEEDTRACKIDX))
					player.display( "Error: seed track #{} not found".format( seedtrack))
				else:
					_LOGGER.debug( "LMSWorker: seed track url= {}".format( seedtrack))
					if config.MNOSEEDDUPS and seedtrack in seedtracks_used:
						_LOGGER.info( "LMSWorker: seed track already used, trying another".format( seedtrack))
						for i in range(0, curr_idx):
							seedtrack_new = player.get_track_fileurl( i)
							if seedtrack_new != None and seedtrack_new not in seedtracks_used:
								seedtrack = seedtrack_new
								break
						_LOGGER.debug( "LMSWorker: final seed track: {}".format( seedtrack))
					seedtracks_used.append( seedtrack)
					try:
						sid = mta.paths.index( seedtrack)
						tracks_added = 0
						stat_tracks = []
						numres_fac = config.MTOPRETFACT
						while tracks_added < config.LMS_MINPLSLEN and numres_fac <= config.MMAXRETFAC:
							_LOGGER.debug( "LMSWorker sim range {} - {}".format( (config.LMS_MINPLSLEN * (numres_fac-1))+1, (config.LMS_MINPLSLEN * numres_fac)+1))
							( rt, rsims) = mus.get_similars( mta.mtracks, mta.mtrackids, sid, ( config.LMS_MINPLSLEN * numres_fac)+1)
							if rt == None:
								break
							for i in range( (config.LMS_MINPLSLEN * (numres_fac-1))+1, len(rt)):
								_LOGGER.debug( "LMSWorker result id: {:3} sim: {:8.6f} file: {}".format( rt[i], rsims[i], mta.paths[ rt[i]]))
								if config.MSKIPDUPS and mta.paths[ rt[i]] in tracks_served: 
									_LOGGER.debug( "LMSWorker: skipped duplicate {} ".format( mta.paths[ rt[i]]))
									continue
								player.playlist_add( mta.paths[ rt[i]])
								tracks_served.append( mta.paths[ rt[i]])
								tracks_added += 1
								stat_tracks.append( { 'path': mta.paths[ rt[i]], 'sim': rsims[i]})
								if tracks_added >= config.LMS_MINPLSLEN:
									break
							numres_fac += 1
						_LOGGER.info( "LMSWorker: added {} tracks (lifetime: {} tracks)".format( tracks_added, mus.get_stat( 'simtracks')))
						player.display( "added {} tracks".format( tracks_added))
						mus.set_stat( 'lastappmsg', "added {} tracks ({})".format( tracks_added, datetime.datetime.now().isoformat(sep=' ', timespec='seconds')))
						mus.set_stat( 'latesttracks', stat_tracks)
						if config.MSKIPDUPS and tracks_added == 0:
							_LOGGER.warn( "LMSWorker: no non-repeated tracks found")
							if curr_idx == 0:
								_LOGGER.info( "LMSWorker: clearing duplicate tracks")
								tracks_served.clear()
							else:
								_LOGGER.info( "LMSWorker: forcing new seed track")
								player.playlist_delete_idx( 0)
						else:
							for i in range( 0, curr_idx):
								_LOGGER.debug( "LMSWorker: delete track #{} ".format( i))
								player.playlist_delete_idx( 0)
					except ValueError:
						_LOGGER.warn("LMSWorker: track {} not found in database".format(seedtrack))
						mus.set_stat( 'lastappmsg', '<span class="text-danger">Error:</span> {} not found in database'.format( seedtrack))
						player.display( "Error: track {} not found".format( seedtrack))

			time.sleep( config.LMS_POLLINTERVAL)

	def __init__(self, lmsserver, musly, mta):
		self._lms = lmsserver
		self._mus = musly
		self._mta = mta
		self._queue = queue.Queue()
		self._event = threading.Event()
		self._thread = threading.Thread(target=self.worker, daemon=True, args=( self._event, self._queue, self._lms, self._mus, self._mta))
		self._mus.set_stat( 'workerstate','paused')
		_LOGGER.debug( "LMSWorker: init done")

	def start(self):
		_LOGGER.debug( "LMSWorker: start")
		self._thread.start()
		
	def pause(self):
		_LOGGER.info( "LMSWorker: pause")
		self._mus.set_stat( 'workerstate','paused')
		self._event.clear()
	def unpause(self):
		_LOGGER.info( "LMSWorker: unpause")
		self._mus.set_stat( 'workerstate','unpaused')
		self._event.set()

	def control(self, msg):
		self._queue.put( msg)

	def join(self):
		self._thread.join()
